/** Minesweeper v1.0.
 * Joe Wingbermuehle 20030712 - 20030713
 */

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;

class Minesweeper {

	private static final String VERSION = "1.0";

	private static JFrame frame;
	private static JPanel gridPanel;

	private static ImageIcon blockIcon;
	private static ImageIcon flagIcon;
	private static ImageIcon bombIcon;
	private static ImageIcon smileIcon;
	private static ImageIcon frownIcon;
	private static ImageIcon coolIcon;
	private static ImageIcon[] uncoveredIcon;
	private static ImageIcon[] ledIcon;

	private static int size = 8;
	private static int mineCount = 10;
	private static int skill = 8;
	private static int clearCount;
	private static int flagsUsed;

	private static final int STATUS_COVERED = 1;
	private static final int STATUS_UNCOVERED = 2;
	private static final int STATUS_FLAGGED = 3;

	private static int[] mines;
	private static int[] status;
	private static JLabel[] buttons;

	private static JLabel[] minesDisplay;
	private static JLabel[] timeDisplay;
	private static JLabel faceDisplay;

	private static Timer timer;

	public static void main(String[] args) {

		blockIcon = new ImageIcon(
			ClassLoader.getSystemResource("icon/block.gif"));
		flagIcon = new ImageIcon(
			ClassLoader.getSystemResource("icon/flag.gif"));
		bombIcon = new ImageIcon(
			ClassLoader.getSystemResource("icon/bomb.gif"));
		smileIcon = new ImageIcon(
			ClassLoader.getSystemResource("icon/smile.gif"));
		frownIcon = new ImageIcon(
			ClassLoader.getSystemResource("icon/frown.gif"));
		coolIcon = new ImageIcon(
			ClassLoader.getSystemResource("icon/cool.gif"));
		uncoveredIcon = new ImageIcon[9];
		for(int x = 0; x <= 8; x++) {
			uncoveredIcon[x] = new ImageIcon(
				ClassLoader.getSystemResource(
				"icon/" + Integer.toString(x) + ".gif"));
		}
		ledIcon = new ImageIcon[10];
		for(int x = 0; x <= 9; x++) {
			ledIcon[x] = new ImageIcon(
				ClassLoader.getSystemResource(
				"icon/led" + Integer.toString(x) + ".gif"));
		}

		timer = new Timer();
		timer.start();

		frame = new JFrame("Minesweeper v" + VERSION);
		frame.getContentPane().setLayout(new BorderLayout());
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

		gridPanel = new JPanel();

		frame.getContentPane().add(gridPanel, BorderLayout.CENTER);

		JMenuBar menuBar = new JMenuBar();

		JMenu gameMenu = new JMenu("Game");
		menuBar.add(gameMenu);
		JMenuItem newItem = new JMenuItem("New Game");
		gameMenu.add(newItem);
		gameMenu.add(new JSeparator());
		JMenuItem exitItem = new JMenuItem("Exit");
		gameMenu.add(exitItem);

		JMenu sizeMenu = new JMenu("Size");
		menuBar.add(sizeMenu);
		ButtonGroup sizeGroup = new ButtonGroup();
		JRadioButtonMenuItem size8Item
			= new JRadioButtonMenuItem("8 x 8");
		sizeMenu.add(size8Item);
		sizeGroup.add(size8Item);
		sizeGroup.setSelected(size8Item.getModel(), true);
		JRadioButtonMenuItem size16Item
			= new JRadioButtonMenuItem("16 x 16");
		sizeMenu.add(size16Item);
		sizeGroup.add(size16Item);
		JRadioButtonMenuItem size32Item
			= new JRadioButtonMenuItem("32 x 32");
		sizeMenu.add(size32Item);
		sizeGroup.add(size32Item);

		JMenu minesMenu = new JMenu("Skill");
		menuBar.add(minesMenu);
		ButtonGroup minesGroup = new ButtonGroup();
		JRadioButtonMenuItem skill1Item
			= new JRadioButtonMenuItem("Novice");
		minesMenu.add(skill1Item);
		minesGroup.add(skill1Item);
		minesGroup.setSelected(skill1Item.getModel(), true);
		JRadioButtonMenuItem skill2Item
			= new JRadioButtonMenuItem("Intermediate");
		minesMenu.add(skill2Item);
		minesGroup.add(skill2Item);
		JRadioButtonMenuItem skill3Item
			= new JRadioButtonMenuItem("Expert");
		minesMenu.add(skill3Item);
		minesGroup.add(skill3Item);

		minesDisplay = new JLabel[3];
		timeDisplay = new JLabel[3];

		JPanel topPanel = new JPanel();
		topPanel.setLayout(new GridLayout(1, 3));

		FlowLayout numberLayout = new FlowLayout(FlowLayout.CENTER, 0, 12);
		JPanel leftPanel = new JPanel();
		leftPanel.setLayout(numberLayout);
		for(int x = 0; x < minesDisplay.length; x++) {
			minesDisplay[x] = new JLabel(ledIcon[0]);
			leftPanel.add(minesDisplay[x]);
		}
		topPanel.add(leftPanel);

		JPanel middlePanel = new JPanel();
		faceDisplay = new JLabel(smileIcon);
		faceDisplay.addMouseListener(new SmileListener());
		middlePanel.add(faceDisplay);
		topPanel.add(middlePanel);

		JPanel rightPanel = new JPanel();
		rightPanel.setLayout(numberLayout);
		for(int x = 0; x < timeDisplay.length; x++) {
			timeDisplay[x] = new JLabel(ledIcon[0]);
			rightPanel.add(timeDisplay[x]);
		}
		topPanel.add(rightPanel);

		frame.getContentPane().add(topPanel, BorderLayout.NORTH);

		frame.setJMenuBar(menuBar);

		createGrid();
		newGame();

		frame.pack();
		frame.setResizable(false);
		frame.show();

		newItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				newGame();
			}
		});

		exitItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				System.exit(0);
			}
		});

		size8Item.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				size = 8;
				createGrid();
				newGame();
				frame.pack();
			}
		});

		size16Item.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				size = 16;
				createGrid();
				newGame();
				frame.pack();
			}
		});

		size32Item.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				size = 32;
				createGrid();
				newGame();
				frame.pack();
			}
		});

		skill1Item.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				skill = 8;
				newGame();
			}
		});

		skill2Item.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				skill = 6;
				newGame();
			}
		});

		skill3Item.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				skill = 4;
				newGame();
			}
		});

	}

	private static class SmileListener extends MouseAdapter {
		public void mouseClicked(MouseEvent e) {
			newGame();
		}
	}

	private static class ButtonListener extends MouseAdapter {

		private int locationX, locationY;

		public ButtonListener(int x, int y) {
			locationX = x;
			locationY = y;
		}

		public void mouseClicked(MouseEvent e) {
			JLabel source = (JLabel)e.getSource();

			int index = locationX + locationY * size;

			if(e.getButton() == MouseEvent.BUTTON1) {
				if(status[index] == STATUS_COVERED) {
					timer.startTimer();
					if(mines[index] > 8) {
						doLose();
					} else {
						clearBlock(locationX, locationY);
					}
				}
			} else if(e.getButton() == MouseEvent.BUTTON3) {
				if(status[index] == STATUS_FLAGGED) {
					status[index] = STATUS_COVERED;
					source.setIcon(blockIcon);
					--flagsUsed;
					showNumber(minesDisplay, mineCount - flagsUsed);
				} else if(status[index] == STATUS_COVERED) {
					if(flagsUsed < mineCount) {
						status[index] = STATUS_FLAGGED;
						source.setIcon(flagIcon);
						++flagsUsed;
						showNumber(minesDisplay, mineCount - flagsUsed);
					}
				}
			}

		}
	}

	private static void newGame() {
		mineCount = (size * size) / skill;
		clearCount = mineCount;
		flagsUsed = 0;

		timer.stopTimer();

		mines = new int[size * size];
		status = new int[size * size];
		for(int x = 0; x < size * size; x++) {
			mines[x] = 0;
			status[x] = STATUS_COVERED;
		}
		for(int x = 0; x < mineCount; x++) {
			int index;
			do {
				index = (int)(Math.random() * (double)(size * size));
			} while(mines[index] == 9);
			mines[index] = 9;
		}
		for(int yc = 0; yc < size; yc++) {
			for(int xc = 0; xc < size; xc++) {
				if(mines[yc * size + xc] > 8) {
					for(int dy = -1; dy <= 1; dy++) {
						if(dy + yc >= 0 && dy + yc < size) {
							for(int dx = -1; dx <= 1; dx++) {
								if(dx + xc >= 0 && dx + xc < size) {
									++mines[(yc + dy) * size + xc + dx];
								}
							}
						}
					}
				}
			}
		}
		for(int x = 0; x < size * size; x++) {
			buttons[x].setIcon(blockIcon);
		}
		faceDisplay.setIcon(smileIcon);
		showNumber(minesDisplay, mineCount);
		showNumber(timeDisplay, 0);
	}

	private static void clearBlock(int xc, int yc) {
		int index = yc * size + xc;
		buttons[index].setIcon(uncoveredIcon[mines[index]]);
		status[index] = STATUS_UNCOVERED;
		++clearCount;
		if(clearCount == size * size) {
			doWin();
		} else if(mines[index] == 0) {
			clearSurrounding(xc, yc);
		}
	}

	private static void clearSurrounding(int xc, int yc) {
		for(int y = -1; y <= 1; y++) {
			if(y + yc >= 0 && y + yc < size) {
				for(int x = -1; x <= 1; x++) {
					if(x + xc >= 0 && x + xc < size) {
						int index = (y + yc) * size + x + xc;
						if(status[index] == STATUS_COVERED) {
							clearBlock(x + xc, y + yc);
						}
					}
				}
			}
		}
	}

	private static void createGrid() {
		buttons = new JLabel[size * size];
		Insets insets = new Insets(0, 0, 0, 0);
		gridPanel.removeAll();
		gridPanel.setLayout(new GridLayout(size, size));
		for(int y = 0; y < size; y++) {
			for(int x = 0; x < size; x++) {
				int index = y * size + x;
				buttons[index] = new JLabel(blockIcon);
				gridPanel.add(buttons[index]);
				buttons[index].addMouseListener(new ButtonListener(x, y));
			}
		}
	}

	private static void doLose() {
		timer.stopTimer();
		for(int x = 0; x < size * size; x++) {
			if(mines[x] > 8) {
				buttons[x].setIcon(bombIcon);
			}
			status[x] = STATUS_UNCOVERED;
		}
		faceDisplay.setIcon(frownIcon);
	}

	private static void doWin() {
		timer.stopTimer();
		for(int x = 0; x < size * size; x++) {
			if(mines[x] > 8) {
				buttons[x].setIcon(flagIcon);
			}
			status[x] = STATUS_UNCOVERED;
		}
		faceDisplay.setIcon(coolIcon);
	}

	private static void showNumber(JLabel[] label, int number) {
		for(int x = label.length - 1; x >= 0; x--) {
			label[x].setIcon(ledIcon[number % 10]);
			number /= 10;
		}
	}

	private static class Timer extends Thread {

		private boolean running = false;
		private int currentTime = 0;

		public void run() {
			for(;;) {
				try {
					Thread.sleep(1000);
				} catch(InterruptedException e) {
					return;
				}
				if(running) {
					updateDisplay();
				}
			}
		}

		public void stopTimer() {
			running = false;
		}

		public void startTimer() {
			if(!running) {
				running = true;
				currentTime = 0;
			}
		}

		public void updateDisplay() {
			if(currentTime == 999) {
				doLose();
			} else {
				++currentTime;
			}
			showNumber(timeDisplay, currentTime);
		}

	}

}


